package org.netkernel.layer0.boot;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.jar.JarFile;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.ZipEntry;

import org.netkernel.container.IKernel;
import org.netkernel.container.ILogger;
import org.netkernel.container.config.IConfiguration;
import org.netkernel.layer0.nkf.impl.NKFContextImpl;
import org.netkernel.layer0.tools.EndpointMonitor;
import org.netkernel.layer0.tools.IEndpointMonitor;
import org.netkernel.layer0.util.Layer0Factory;
import org.netkernel.layer0.util.SpaceClassLoader;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLReadable;
import org.netkernel.urii.impl.NetKernelException;
import org.xml.sax.SAXException;


public class BootUtils
{
	public static String TEMP_FILE_BASE="netkernel_";
	
	public static String PROPERTY_NETKERNEL_INSTALL="netkernel.install.path";
	
	private static FileFilter sDynamicFileFilter;
	static
	{	sDynamicFileFilter = new FileFilter()
		{	public boolean accept(File aFile)
			{	boolean result=false;
				if (!aFile.isHidden() && (aFile.isDirectory() || aFile.getName().endsWith(".class")))
				{	result=true;
				}
				return result;
			}
		};
	}
	
	
	/** Attempt to build a module from the given URI using the given factories
	 * @param aModuleURI the source for the module
	 * @param aKernel a reference to the kernel the module will be deployed to
	 * @param aFactories an array of factories used in supplied order
	 * @return the module
	 * @throws Exception if there is a problem building a module or no factory is suitable to build module
	 */
	public static IModule buildModule(URI aModuleURI, IKernel aKernel, IModuleFactory[] aFactories) throws Exception
	{ IModule result=null;
		for (int i=0; i<aFactories.length && result==null; i++)
		{	result = aFactories[i].buildModule(aModuleURI,aKernel);
		}
		if (result==null)
		{	
			String moduleURIString=fixURIString(aModuleURI.toString());
			URL moduleConfig = new URL(moduleURIString);
			URLConnection c = moduleConfig.openConnection();
			NetKernelException e2;
			try
			{	c.getInputStream();
				e2= new NetKernelException(aKernel.getLogger().format("EX_BOOT_NO_FACTORY",aModuleURI));
			}
			catch (Exception e)
			{	e2= new NetKernelException(aKernel.getLogger().format("EX_BOOT_NO_FACTORY",aModuleURI),null,e);
			}
			throw e2;
		}
		return result;
	}
	
	public static void flushExpandedJars(IConfiguration aConfig)
	{
		String expandDirString=aConfig.getString("netkernel.layer0.expandDir",null);
		if (expandDirString==null)
		{	// delete these temp files
			File tempDir = new File(System.getProperty("java.io.tmpdir"));
			FileFilter ff = new FileFilter()
			{	public boolean accept(File aFile)
				{	return aFile.getName().startsWith(TEMP_FILE_BASE);
				}
			};
			File[] filenames = tempDir.listFiles(ff);
			for (int i=0; i<filenames.length; i++)
			{	filenames[i].delete();
			}
		}
	}
	
	
	public static XMLReadable getXMLFrom(URI aBase, String aResource) throws IOException, SAXException
	{	String moduleURIString=fixURIString(aBase.toString());
		URL moduleConfig = new URL(moduleURIString+aResource);
		URLConnection c = moduleConfig.openConnection();
		InputStream is = c.getInputStream();
		return new XMLReadable(is);
	}
	
	public static URI findWildCardMatch(URI aBase, String aRegex) throws IOException
	{
		String moduleURIString=fixURIString(aBase.toString());
		URL moduleURL = new URL(moduleURIString); 
		String scheme=moduleURL.getProtocol();
		URI result=null;
		
		if (scheme.equals("file"))
		{
			File dir = new File(aBase);
			
		}
		
		return result;
	}
	
	
	/** Creates a classloader from a base URI with optional inclusion of a sub directory of jar resources
	 * @param aBase base URI for classloader
	 * @param aParent parent classloader
	 * @param aLibDir null or a relative path to directory containing jars to put into classloader (don't end with a slash)
	 * @param aExports list of regex strings to put into classloader
	 * @param aToString debug string that is returned on toString() method in classloader
	 * @param aLayer0Config config used to find netkernel.layer0.expandDir if needed
	 * @param aLogger logger to log messages to
	 * @return the created classloader
	 * @throws IOException throws if there are problems creating classloader
	 */
	public static SpaceClassLoader getClassLoaderFrom(URI aBase, ClassLoader aParent, String aLibDir, List aExports, String aToString, IKernel aKernel ) throws IOException
	{
		List jarURLs = new ArrayList();
		
		String moduleURIString=fixURIString(aBase.toString());
		URL moduleURL = new URL(moduleURIString); 
		jarURLs.add( moduleURIString );
		
		if (aLibDir!=null)
		{
			String scheme=moduleURL.getProtocol();
			if (scheme.equals("jar"))
			{	
				/*
				String expandDirString=aLayer0Config.getString("netkernel.layer0.expandDir",null);
				File expandDir=null;
				if (expandDirString!=null && expandDirString.length()>0)
				{	expandDir=new File(expandDirString);
				}
				*/	
				
				JarURLConnection juc = (JarURLConnection)moduleURL.openConnection();
				JarFile jf = juc.getJarFile();
				Matcher m = Pattern.compile(aLibDir+"/[^/]*\\.jar").matcher(""); // any jar directly in the lib dir
				for (Enumeration e = jf.entries(); e.hasMoreElements(); )
				{	ZipEntry entry = (ZipEntry)e.nextElement();
					m.reset(entry.getName());
					if (m.matches())
					{	String url = expandNestedJar(jf,juc.getLastModified(),entry,aLibDir,aKernel, aBase.toString());
						jarURLs.add(url);
					}
				}
			}
			else if (scheme.equals("file"))
			{	File libDir = new File(aBase);
				libDir=new File(libDir,aLibDir);
				if (libDir.exists())
				{
					FileFilter ff = new FileFilter()
					{	public boolean accept(File aFile)
						{	return aFile.getName().endsWith(".jar");
						}
					};
					File[] filenames = libDir.listFiles(ff);
					for (int i=0; i<filenames.length; i++)
					{	jarURLs.add( "jar:"+Utils.fileAsURIString(filenames[i])+"!/" );
					}
				}
			}
			else
			{	throw new IllegalArgumentException(aKernel.getLogger().format("EX_BOOT_BAD_SCHEME",scheme));
			}
		}
		return new SpaceClassLoader(aParent,jarURLs, aExports, aToString);
	}
	
	public static String expandNestedJar(JarFile aJF, long aJarLastModified, ZipEntry aEntry, String aLibDir, IKernel aKernel, String moduleURI) throws IOException
	{
		String expandDirString=aKernel.getConfiguration().getString("netkernel.layer0.expandDir",null);
		File expandDir=null;
		if (expandDirString!=null && expandDirString.length()>0)
		{	expandDir=new File(expandDirString);
		}	
		String name=aEntry.getName();
		int i=name.lastIndexOf('.');
		name=name.substring(aLibDir.length()+1,i);
		File expanded=getUniqueFile(name,expandDir, moduleURI);
		boolean skipExpansion=false;
		if (expandDir!=null && expanded.exists())
		{	//System.out.println(new Date(aJarLastModified)+" > "+new Date(expanded.lastModified()));
			if (aJarLastModified<expanded.lastModified())
			{	skipExpansion=true;
			}
		}
		if (!skipExpansion)
		{	aKernel.getLogger().log(ILogger.INFO,BootUtils.class,"EX_BOOT_EXPAND",name,expanded.getAbsolutePath());
			FileOutputStream fos = new FileOutputStream(expanded);
			InputStream is=aJF.getInputStream(aEntry);
			Utils.pipe(is,fos);
		}
		String url = "jar:"+Utils.fileAsURIString(expanded)+"!/";
		return url;
	}
	
	public static String fixURIString(String aURIString)
	{
		if(File.separatorChar=='\\' && !aURIString.startsWith("file:///"))
		{	aURIString="file:///"+aURIString.substring(6);
		}
		if (aURIString.endsWith(".jar"))
		{	aURIString = "jar:"+aURIString+"!/";
		}
		return aURIString;
	}
	
	private static File getUniqueFile(String aName, File aDestDir, String moduleURI) throws IOException
	{
		String mod=moduleURI.substring(moduleURI.lastIndexOf("/")+1);
		mod=mod.replaceAll("\\.s?jar", "___");
		String prefix=mod+aName;
		StringBuffer sb = new StringBuffer(prefix.length()+4);
		for (int n=0; n<prefix.length(); n++)
		{	char c = prefix.charAt(n);
			if (Character.isLetterOrDigit(c) || c=='-' || c=='.')
			{	sb.append(c);
			}
			else
			{	sb.append('_');
			}
		}
		prefix = sb.toString();		
		
		File dest;
		if (aDestDir!=null)
		{	dest = new File(aDestDir,prefix+".jar");
		}
		else
		{	dest = File.createTempFile(TEMP_FILE_BASE+prefix+"_",".jar");
			//dest.deleteOnExit();
			// delete on exit doesn't work because file is locked by classloader!
		}
		return dest;
	}
	
	public static List getJavaClasses(File aBase)
	{	List l = new ArrayList(256);
		recurseDir(aBase,l);
		return l;
	}
	
	public static boolean isModifiedSince(long aTimestamp, File[] aFiles)
	{	for (int i=0; i<aFiles.length; i++)
		{	File f= aFiles[i];
			if (f.lastModified()>aTimestamp)
			{
				if (f.isDirectory())
				{	File[] files = f.listFiles(sDynamicFileFilter);
					if (files != null)
					{	for (int j=0; j<files.length; j++)
						{	File f2=files[j];
							if (f2.getName().endsWith(".class"))
							{	if (f2.lastModified()>aTimestamp)
								{	return true;
								}
							}
						}
					}
				}
				else
				{	return true;
				}
			}
		}
		return false;	
	}
	
	private static void recurseDir(File aDir, List aResult)
	{	
		
		File[] files = aDir.listFiles(sDynamicFileFilter);
		if (files != null)
		{	
			for (int i=0; i<files.length; i++)
			{	File f = files[i];
				
				boolean isDirectory = f.isDirectory();
				String name = f.getName();
				
				if (isDirectory)
				{	aResult.add(f);	
					recurseDir(f,aResult);
				}
				else
				{	if (name.endsWith(".class"))
					{	aResult.add(f);
					}
				}
			}
		}
	}
	
	public static boolean bootedFromJar(IConfiguration config)
	{	boolean inJar = getInstallPath(config)==null;
		return inJar;
	}
	
	public static String getBootedJarName()
	{	String classpath=System.getProperty("java.class.path");
		int j=classpath.indexOf(File.pathSeparatorChar);
		if (j>0)
		{	classpath=classpath.substring(0,j);
		}
		return classpath;
	}
	
	public static String getInstallPath(IConfiguration config)
	{	String path=null;
		try
		{	path=config.getString(PROPERTY_NETKERNEL_INSTALL);
		}
		catch(Exception e) { /*Do nothing */}
		return path;
	}
	
	public static URI getInitModules(IConfiguration config)
	{	String modstr=null;
		URI moduleListURI=null;
		try
		{	modstr=config.getString("netkernel.init.modules");
		}
		catch(Exception e){ /*Do nothing */}
		if(modstr!=null && modstr.length()>0)
		{	URI base=URI.create(BootUtils.getInstallPath(config));
			URI modulesURI=URI.create(modstr);
			if(!modulesURI.isAbsolute())
			{	moduleListURI=base.resolve(modulesURI);
			}
			else
			{	moduleListURI=modulesURI;
			}
		}
		return moduleListURI;
	}
	
	public static void cleanupOnHotRestart()
	{	IEndpointMonitor monitor=Layer0Factory.getMonitorInstance();
		if (monitor!=null)
		{	monitor.clear();
		}
		NKFContextImpl.cleanup();
		
		//restore context classloaders for all threads
		ThreadGroup rootThreadGroup = Thread.currentThread( ).getThreadGroup( );
	    ThreadGroup ptg;
	    while ( (ptg = rootThreadGroup.getParent( )) != null )
	    {	rootThreadGroup = ptg;
	    }
	    Thread[] threads = new Thread[rootThreadGroup.activeCount()*2];
	    rootThreadGroup.enumerate(threads);
		for (Thread thread : threads)
		{	if (thread!=null)
			{	thread.setContextClassLoader(Class.class.getClassLoader());
			}
			else
			{	break;
			}
		}
	}
}
