/******************************************************************************
  (c) Copyright 2002-2006, 1060 Research Ltd                                   

  This Software is licensed to You, the licensee, for use under the terms of   
  the 1060 Public License v1.0. Please read and agree to the 1060 Public       
  License v1.0 [www.1060research.com/license] before using or redistributing   
  this software.                                                               

  In summary the 1060 Public license has the following conditions.             
  A. You may use the Software free of charge provided you agree to the terms   
  laid out in the 1060 Public License v1.0                                     
  B. You are only permitted to use the Software with components or applications
  that provide you with OSI Certified Open Source Code [www.opensource.org], or
  for which licensing has been approved by 1060 Research Limited.              
  You may write your own software for execution by this Software provided any  
  distribution of your software with this Software complies with terms set out 
  in section 2 of the 1060 Public License v1.0                                 
  C. You may redistribute the Software provided you comply with the terms of   
  the 1060 Public License v1.0 and that no warranty is implied or given.       
  D. If you find you are unable to comply with this license you may seek to    
  obtain an alternative license from 1060 Research Limited by contacting       
  license@1060research.com or by visiting www.1060research.com                 

  NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC  
  LICENSE V1.0 FOR DETAILS                                                     

  THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ     
  THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license      

  File:          $RCSfile$
  Version:       $Name$ $Revision$
  Last Modified: $Date$
 *****************************************************************************/
package org.netkernel.layer0.boot;

import java.net.URI;
import java.util.List;

import org.netkernel.container.IKernel;
import org.netkernel.layer0.meta.IModuleMeta;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.ISpaceWithIdentity;

/** Interface for "new" modules
 * Module lifecycle:<ul>
 * <li>a) constructor should construct spaces and register them with kernel</li>
 * <li>b) linkModule() is called and module should call linkSpace() for all spaces it has</li>  
 * <li>b) commissionModule() should call commissionSpace() for all spaces it has</li>
 * <li>c) <do work here- module is live>
 * <li>d) getRegisteredSpaces() should return an array of all spaces that are registered with kernel</li>
 * <li>e) decommissionModule() should call deCommissionSpace() for all spaces it has and deregister them with kernel</li>
 * </ul>
 * @author Tony Butterfield
 */
public interface IModule
{
	/** Called to allow spaces within a module to link to other spaces via a kernel lookup
	 *  using IKernel.getSpace(IIdentifier, Version, Version)
	 * @param aKernel
	 * @throws Exception
	 */
	void linkModule(IKernel aKernel) throws Exception;
	
	/**  Called to allow module specific initialisation in particular instantiating any spaces
	 * @param aKernel the kernel instance this module is being commisioned into
	 * @throws Exception thrown for any implementation specific failures that may occur
	 */
	void commissionModule(IKernel aKernel) throws Exception;
	
	/** Called to allow module to perform any processing that may involve issuing requests. This
	 *  will happen before any external transport requests enter system
	 * @throws Exception
	 */
	void postCommissionModule() throws Exception;
	
	/** Called to allow module to perform any last chance processing that may involve issuing requests
	 * before it is decommissioned. All external transport requests will be completed or held whilst this
	 * is happening
	 * @throws Exception
	 */
	void preDecommissionModule() throws Exception;
	
	/** Called to allow module specific decommisioning behaviour - usually nothing is needed,
	 * this method is also called if the module fails to link or commission without error.
	 * @param aKernel
	 * @throws Exception thrown for any implementation specific failures that may occur
	 */
	void decommisionModule(IKernel aKernel) throws Exception;
	
	/** Returns the list of all public spaces registered with the kernel owned by this module
	 * @return
	 */
	ISpaceWithIdentity[] getPublicSpaces();
	
	/** Returns the list of all spaces owned by this module
	 */
	ISpace[] getSpaces();
	
	/** A module may implement ability to detect changes which will require it to be decommissioned and restarted
	 */
	boolean isExpired(long aNow);
	
	/** Determine if this module should be expired and hence reloaded when the system is restarted.
	 * This is the case if a module has loaded classes which could change when new modules are introduced to
	 * the system.
	 * @param aNewModules the new modules that are being introduced
	 * @return true if the module should be reloaded.
	 */
	boolean isExpiredOnRestartWithNewModules(List<IModule> aNewModules);
	
	/** Return the source URI where the module has been instantiated from
	 */
	URI getSource();
	
	/** Return the modules meta data
	 */
	IModuleMeta getMeta();
}