package org.netkernel.layer0.util;

import org.netkernel.container.IMessages;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.representation.IBinaryStreamRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.representation.impl.HDSFactory;
import org.netkernel.layer0.util.RequestBuilder.Arguments.ArgType;
import org.netkernel.request.impl.RequestFactory;
import org.netkernel.urii.INetKernelException;
import org.netkernel.urii.impl.NetKernelException;
import org.w3c.dom.*;

import java.lang.reflect.Constructor;
import java.util.*;

import javax.xml.parsers.ParserConfigurationException;

/** Declarative Request construction factory. Builds a request from standard syntax.
 * @author tab
 *
 */
public class RequestBuilder
{
	private static final PairList VALIDATION = new PairList(5);
	static
	{	VALIDATION.put("./*[name()!='identifier' and name()!='argument' and name()!='verb' and name()!='representation' and name()!='varargs']", "MSG_RQT_BUILDER1");
		VALIDATION.put("self::node()[count( identifier )!=1]", "MSG_RQT_BUILDER2");
		VALIDATION.put("./argument/attribute::*[name()!='name' and name()!='method' and name()!='tolerant']", "MSG_RQT_BUILDER3");
		VALIDATION.put("./argument[not(@name)]", "MSG_RQT_BUILDER3");
		VALIDATION.put("./*[string-length(text())=0 and name()!='argument' and name()!='varargs']", "MSG_RQT_BUILDER4");
		VALIDATION.put("./argument[@name='primary' and @method]", "MSG_RQT_BUILDER5");
		VALIDATION.put("./argument[@method!='value' and @method!='data-uri' and @method!='as-string' and @method!='from-string']", "MSG_RQT_BUILDER6");
		VALIDATION.put("./argument[request and @method]", "MSG_RQT_BUILDER7");
		VALIDATION.put("./argument[* and @method!='data-uri']", "MSG_RQT_BUILDER8");
	}
	private static String sNoArgExceptionId;
	
	private int mVerb;
	private String mIdentifier;
	private String mRepClass;
	private boolean mVarargs;
	private Arg[] mArguments;
	private boolean mPassthroughRepresentation;
	
	private static class Arg
	{	public Arg(String aName, String aMethod, Object aValue, boolean aTolerant)
		{	mName=aName;
			mMethod=aMethod;
			mValue=aValue;
			mTolerant=aTolerant;
		}
		public String mName;
		public String mMethod;
		public Object mValue;
		public boolean mTolerant;
	}

	
	/** Create a request builder from DOM XML syntax
	 * @param aRequestDeclaration a document containing the request declaration
	 * @param aMessages kernel messages interface to find error messages
	 * @throws INetKernelException thrown if request declaration is malformed
	 */
	public RequestBuilder(Node aRequestDeclaration, IMessages aMessages) throws INetKernelException
	{
		//validate
		Element root;
		switch(aRequestDeclaration.getNodeType())
		{	case Node.ELEMENT_NODE:
				root=(Element)aRequestDeclaration;
				break;
			case Node.DOCUMENT_NODE:
				root=((Document)aRequestDeclaration).getDocumentElement();
				break;
			default:
				throw Utils.createFormattedException("EX_RQT_BUILDER_CONFIG", "MSG_RQT_BUILDER0",aMessages);
		}
		PairList validation=FastSchematron.validate(root, VALIDATION);
		if (validation.size()>0)
		{	String m = FastSchematron.formatValidationError(validation,aMessages);			
			throw Utils.createFormattedException("EX_RQT_BUILDER_CONFIG", "MSG_RAW",aMessages, m);
		}
		
		XMLReadable r = new XMLReadable(root);
		mIdentifier=r.getText("identifier");
		String verbString=r.getText("verb");
		if (verbString.length()>0)
		{	mVerb=RequestFactory.parseVerb(verbString);
		}
		mRepClass=r.getText("representation");
		mVarargs=r.getFirstNode("varargs")!=null;
		
		List<Node> argNodes=r.getNodes("argument");
		mArguments=new Arg[argNodes.size()];
		int j=0;
		for (Iterator<Node> i=argNodes.iterator(); i.hasNext(); )
		{	Element argNode = (Element)i.next();
			String name=argNode.getAttribute("name");
			String method=argNode.getAttribute("method");
			Object value;
			boolean tolerant=argNode.getAttribute("tolerant").equals("true");
			Node literalRoot = XMLUtils.getFirstChildElement(argNode);
			if (literalRoot==null)
			{	value=XMLUtils.getText(argNode).trim();
			}
			else if (literalRoot.getNodeName().equals("request"))
			{	RequestBuilder nested=new RequestBuilder(literalRoot,aMessages);
				value=nested;
			}
			else
			{	value=literalRoot;
			}
			Arg arg=new Arg(name,method,value,tolerant);
			mArguments[j++]=arg;
		}		
	}
	
	/** By default a declarative request with have a requested representation type of java.lang.Object.
	 * If this PassthroughRequestedRepresentation is set to true then the requested representation
	 * from the supplied INKFRequestContext is used when buildRequest() is called.
	 * @param aPassthrough
	 */
	public void setPassthroughRequestedRepresentation(boolean aPassthrough)
	{	mPassthroughRepresentation=aPassthrough;
	}
	
	public void writeHDS(HDSBuilder b)
	{
		b.pushNode("request");
		if (mVerb!=0) b.addNode("verb", RequestFactory.verbString(mVerb));
		b.addNode("identifier", mIdentifier);
		if (mArguments.length>0)
		{	b.pushNode("arguments");
			for (Arg arg: mArguments)
			{	String v=null;
				Object value=arg.mValue;
				if (value instanceof String)
				{	v=(String)value;
				}
				b.pushNode("arg",v);
				b.addNode("@name", arg.mName);
				if (arg.mMethod.length()>0) b.addNode("@method", arg.mMethod);
				if (arg.mTolerant) b.addNode("@tolerant", arg.mTolerant);
				if (value instanceof RequestBuilder)
				{	((RequestBuilder)value).writeHDS(b);
				}
				else if (!(value instanceof String))
				{	b.addNode("literal", null);
				}
				b.popNode();
			}
			b.popNode();
		}
		if (mVarargs) b.addNode("varargs", null);
		if (mRepClass.length()>0) b.addNode("representation", mRepClass);
		b.popNode();
	}
	
	/** Build a request instance from this builder
	 * @param aContext the INKFRequestContext in which to build the request
	 * @param aArguments a set of preset arguments that may be referenced on the request, may be null
	 * @param aRepresentationClassLoader the classloader to load the representation, if null a classloader based on the current
	 * space will be used.
	 * @return the built request
	 * @throws Exception
	 */
	public INKFRequest buildRequest(INKFRequestContext aContext, Arguments aArguments, ClassLoader aRepresentationClassLoader) throws Exception
	{
		String identifier=buildArgumentValueWithSubs(mIdentifier,aContext);
		INKFRequest result=aContext.createRequest(identifier);
		
		if (mVerb!=0)
		{	result.setVerb(mVerb);
		}
		
		String representationString=mRepClass;
		if (representationString.length()>0)
		{	if (aRepresentationClassLoader==null)
			{	aRepresentationClassLoader=new RequestScopeClassLoader(aContext.getKernelContext().getThisKernelRequest().getRequestScope());
			}
			Class representationClass=aRepresentationClassLoader.loadClass(representationString);
			result.setRepresentationClass(representationClass);
		}
		else if (mPassthroughRepresentation)
		{	result.setRepresentationClass(aContext.getThisRequest().getRepresentationClass());
		}
		
		for (Arg arg: mArguments )
		{	
			String name=arg.mName;
			
			String method=arg.mMethod;
			boolean isPrimaryArg=name.equals("primary");
			
			Object v = arg.mValue;
			if (v instanceof String)
			{	String value=(String)v;
			
				try
				{
					value=buildArgumentValueWithSubs(value,aContext);
					
					Object valueRepresentation=null;
					ArgType valueType=null;
					if (aArguments!=null)
					{	TupleList argValues=aArguments.getData();
						for (int j=0; j<argValues.size(); j++)
						{	if (value.equals(argValues.getValue(j,0)))
							{	
								valueType=(ArgType)argValues.getValue(j,1);
								if (valueType==ArgType.REFERENCE)
								{	value=(String)argValues.getValue(j,2);
								}
								else
								{	valueRepresentation=argValues.getValue(j,2);
								}
								break;
							}
						}
					}
					
					if (method.equals("value") && valueRepresentation==null)
					{	valueRepresentation = aContext.sourceForResponse(value);
						valueType=ArgType.RESPONSE;
					}
					else if (method.equals("data-uri"))
					{	
						INKFResponseReadOnly<IBinaryStreamRepresentation> resp;
						if (valueRepresentation==null)
						{	resp=aContext.sourceForResponse(value, IBinaryStreamRepresentation.class);
						}
						else
						{	Object source;
							if (valueType==ArgType.RESPONSE)
							{	source=((INKFResponseReadOnly)valueRepresentation).getRepresentation();
							}
							else
							{	source=valueRepresentation;
							}
							resp=aContext.transreptForResponse(source, IBinaryStreamRepresentation.class);
						}
						
						value=Utils.toDataURI(resp.getRepresentation(), resp.getMimeType());
						valueRepresentation=null;
					}
					else if (method.equals("as-string"))
					{
						if (value.startsWith("arg:"))
						{	String argName=value.substring(4);
							String argValue=aContext.getThisRequest().getArgumentValue(argName);
							valueRepresentation=argValue;
						}
						else
						{	valueRepresentation=value;
						}
						valueType=ArgType.VALUE;
					}
					else if (method.equals("from-string"))
					{
						if (valueRepresentation!=null)
						{	if (valueRepresentation instanceof String)
							{	value=(String)valueRepresentation;
							}
							else
							{	value=aContext.transrept(valueRepresentation, String.class);
							}
						}
						else
						{	value=aContext.source(value,String.class);
						}
						valueRepresentation=null;
					}
					
					
					if (valueRepresentation==null)
					{	try
						{	if (isPrimaryArg)
							{	INKFResponseReadOnly sourcedPrimary=aContext.sourceForResponse(value);
								result.addPrimaryArgumentFromResponse(sourcedPrimary);
							}
							else
							{	result.addArgument(name, value);
							}
						}
						catch (NKFException e)
						{	//ignore error if argument not found, tolerant omission
							if (!value.startsWith("arg:"))
							{	throw e;
							}
							else
							{	if (sNoArgExceptionId==null)
								{	sNoArgExceptionId=aContext.getKernelContext().getKernel().getLogger().format("EX_NKF_REQ_NO_ARG");
								}
								if (!e.getDeepestId().equals(sNoArgExceptionId))
								{	throw e;
								}
							}
						}
					}
					else if (valueType==ArgType.RESPONSE)
					{	INKFResponseReadOnly valueResponse=(INKFResponseReadOnly)valueRepresentation;
						if (isPrimaryArg)
						{	result.addPrimaryArgumentFromResponse(valueResponse);						
						}
						else
						{	result.addArgumentFromResponse(name,valueResponse);
						}
					}
					else if (valueType==ArgType.REFERENCE)
					{	if (isPrimaryArg)
						{	result.addPrimaryArgument(valueRepresentation);
						}
						else
						{	result.addArgument(name,(String)valueRepresentation);
						}
					}
					else
					{	if (isPrimaryArg)
						{	result.addPrimaryArgument(valueRepresentation);
						}
						else
						{	result.addArgumentByValue(name, valueRepresentation);
						}
					}
				}
				catch(Exception e)
				{	if (!arg.mTolerant)
					{	throw e;
					}
				}
			}
			else if (v instanceof RequestBuilder)
			{	RequestBuilder b = (RequestBuilder)v;
				INKFRequest req=b.buildRequest(aContext, aArguments, aRepresentationClassLoader);
				result.addArgumentByRequest(name, req);
			}
			else
			{	Node literalRoot=(Node)v;
				Object[] argTypePair = parseLiteral(literalRoot, aRepresentationClassLoader,aContext);
				Object value=argTypePair[0];
				
				if (method.equals("data-uri"))
				{	IBinaryStreamRepresentation rep=aContext.transrept(value, IBinaryStreamRepresentation.class);
					String value2=Utils.toDataURI(rep, "text/xml");
					result.addArgument(name,value2);
				}
				else
				{	if (isPrimaryArg)
					{	result.addPrimaryArgument(value);
					}
					else
					{	result.addArgumentByValue(name,value);
					}
				}
			}
		}
		
		//varargs
		if (mVarargs)
		{	INKFRequestReadOnly reqIn=aContext.getThisRequest();
			for (int i=0; i<reqIn.getArgumentCount(); i++)
			{
				String name=reqIn.getArgumentName(i);
				boolean found=false;
				for (int j=0; j<mArguments.length; j++)
				{	if (mArguments[j].mName.equals(name))
					{	found=true;
						break;
					}
				}
				if (!found)
				{	String value=reqIn.getArgumentValue(i);
					try
					{	result.addArgumentAsAbsolute(name, value);
					}
					catch (Exception e)
					{	//tolerant of failure to add
					}
				}
			}
		}
		
		return result;
	}
	
	private static String buildArgumentValueWithSubs(String aArgValue, INKFRequestContext aContext) throws Exception
	{	int p1;
		while((p1=aArgValue.indexOf("[[arg:"))>=0)
		{	int p2 = aArgValue.indexOf("]]");
			if (p2>0)
			{	String argName = aArgValue.substring(p1+6,p2);
				String argValue=aContext.getThisRequest().getArgumentValue(argName);
				if (argValue!=null)
				{	aArgValue = aArgValue.substring(0,p1)+argValue+aArgValue.substring(p2+2);
				}
				else
				{	IMessages messages= aContext.getKernelContext().getKernel().getLogger();
					throw Utils.createFormattedNKFException("EX_NKF_REQ_NO_ARG", "MSG_NKF_REQ_NO_ARG", null, null, messages, argName);
				}
			}
			else
			{	break;
			}
		}
		return aArgValue;
	}
	
	private enum LiteralType { BYTE, SHORT, INTEGER, LONG, FLOAT, DOUBLE, BOOLEAN, CHAR, STRING, XML, HDS, OTHER };
	
	public static Object[] parseLiteral(Node aNode, ClassLoader aRepresentationClassLoader, INKFRequestContext aContext) throws Exception
	{
		Object result;
		Class resultType;
		if (aNode.getNodeName().equals("literal") && aNode.getNodeType()==Node.ELEMENT_NODE && ((Element)aNode).getAttributeNode("type")!=null)
		{	Element e=(Element)aNode;
			String type=e.getAttribute("type");
			LiteralType lt;
			try
			{	lt=LiteralType.valueOf(type.toUpperCase());
			}
			catch (IllegalArgumentException e2)
			{	lt=LiteralType.OTHER;
			}
			if (lt!=LiteralType.XML && lt!=LiteralType.HDS && lt!=LiteralType.OTHER && XMLUtils.getFirstChildElement(e)!=null)
			{	throw aContext.createFormattedException("EX_RQT_BUILDER_CONFIG", "MSG_RQT_LITERAL_FRAG_UNEXPECTED", XMLUtils.getPathFor(aNode), null,type);
			}
			String textValue=XMLUtils.getText(e);
			switch (lt)
			{
			case STRING:
				result = textValue;
				resultType=String.class;
				break;
			case BYTE:
				result = Byte.parseByte(textValue);
				resultType=byte.class;
				break;
			case SHORT:
				result = Short.parseShort(textValue);
				resultType=short.class;
				break;
			case INTEGER:
				result = Integer.parseInt(textValue);
				resultType=int.class;
				break;
			case LONG:
				result = Long.parseLong(textValue);
				resultType=long.class;
				break;
			case FLOAT:
				result = Float.parseFloat(textValue);
				resultType=float.class;
				break;
			case DOUBLE:
				result = Double.parseDouble(textValue);
				resultType=double.class;
				break;
			case BOOLEAN:
				result = Boolean.parseBoolean(textValue);
				resultType=boolean.class;
				break;
			case CHAR:
				if (textValue.length()!=1)
				{	throw aContext.createFormattedException("EX_RQT_BUILDER_CONFIG", "MSG_RQT_LITERAL_CHAR", XMLUtils.getPathFor(aNode), null);
				}
				result = textValue.charAt(0);
				resultType=char.class;
				break;
			case XML:
				Element fragmentRoot=XMLUtils.getFirstChildElement(e);
				if (fragmentRoot==null || XMLUtils.getNextSiblingElement(fragmentRoot)!=null)
				{	throw aContext.createFormattedException("EX_RQT_BUILDER_CONFIG", "MSG_RQT_LITERAL_XML", XMLUtils.getPathFor(aNode), null);
				}
				Document fragment=XMLUtils.newDocument();
				fragment.appendChild(fragment.importNode(fragmentRoot, true));
				result=fragment;
				resultType=Document.class;
				break;
			case HDS:
				Element fragmentRoot2=XMLUtils.getFirstChildElement(e);
				if (fragmentRoot2==null)
				{	throw aContext.createFormattedException("EX_RQT_BUILDER_CONFIG", "MSG_RQT_LITERAL_HDS", XMLUtils.getPathFor(aNode), null);
				}
				result=HDSFactory.parseDOM(fragmentRoot2);
				resultType=IHDSNode.class;
				fragmentRoot2=XMLUtils.getNextSiblingElement(fragmentRoot2);
				if (fragmentRoot2!=null)
				{	//forest not tree
					HDSBuilder b = new HDSBuilder();
					b.importChildren((IHDSNode)result);
					do
					{	IHDSNode n=HDSFactory.parseDOM(fragmentRoot2);
						b.importChildren(n);
						fragmentRoot2=XMLUtils.getNextSiblingElement(fragmentRoot2);
					} while(fragmentRoot2!=null);
					result=b.getRoot();
				}
				break;
				
			default:
				if (aRepresentationClassLoader==null)
				{	aRepresentationClassLoader=new RequestScopeClassLoader(aContext.getKernelContext().getThisKernelRequest().getRequestScope());
				}
				Class c;
				try
				{	c = aRepresentationClassLoader.loadClass(type);
				}
				catch (ClassNotFoundException ex)
				{	throw aContext.createFormattedException("EX_RQT_BUILDER_CONFIG", "MSG_RQT_LITERAL_CLASS", XMLUtils.getPathFor(aNode), null,type);
				}
				List<Object[]> args = new ArrayList();
				for (Element e2=XMLUtils.getFirstChildElement(e); e2!=null; e2=XMLUtils.getNextSiblingElement(e2))
				{	if (!e2.getNodeName().equals("literal"))
					{	throw aContext.createFormattedException("EX_RQT_BUILDER_CONFIG", "MSG_RQT_LITERAL_CONSTRUCTOR_ARG", XMLUtils.getPathFor(e2), null);
					}
					else
					{	Object[] argTypePair=parseLiteral(e2,aRepresentationClassLoader,aContext);
						args.add(argTypePair);
					}
				}
				Class[] types = new Class[args.size()];
				for (int i=0; i<types.length; i++)
				{	Object[] argTypePair=args.get(i);
					types[i]=(Class)argTypePair[1];
				}
				Constructor constructor;
				try
				{	 constructor = c.getConstructor(types);
				}
				catch (NoSuchMethodException ex)
				{	throw aContext.createFormattedException("EX_RQT_BUILDER_CONFIG", "MSG_RQT_LITERAL_CONSTRUCTOR_NOT_FOUND", XMLUtils.getPathFor(aNode), null,c.getName(),Utils.arrayToString(types, 128));
				}
				try
				{	Object[] argArray=new Object[args.size()];
					for (int i=0; i<argArray.length; i++)
					{	Object[] argTypePair=args.get(i);
						argArray[i]=argTypePair[0];
					}
					result=constructor.newInstance(argArray);
					resultType=c;
				}
				catch (Exception ex)
				{	throw ex;
				}
			}
		}
		else
		{	//treat as regular xml literal
			if (aNode==aNode.getOwnerDocument().getDocumentElement())
			{	result=aNode.getOwnerDocument();
			}
			else
			{	Document fragment=XMLUtils.newDocument();
				fragment.appendChild(fragment.importNode(aNode, true));
				result=fragment;
			}
			resultType=Document.class;
		}
			
		return new Object[] { result, resultType };
	}
	
	public static class Arguments
	{
		private TupleList mArguments;
		enum ArgType { REFERENCE, VALUE, RESPONSE };
		
		public Arguments()
		{	mArguments=new TupleList(3,4);
		}
		
		public void addArgumentByValue(String aName, Object aRepresentation)
		{	mArguments.put(aName,ArgType.VALUE,aRepresentation);
		}
		
		public void addArgument(String aName, String aIdentifier)
		{	mArguments.put(aName,ArgType.REFERENCE,aIdentifier);
		}
		
		public void addArgumentFromResponse(String aName,  INKFResponseReadOnly aResponse)
		{	mArguments.put(aName,ArgType.RESPONSE,aResponse);
		}
		
		TupleList getData()
		{	return mArguments;
		}
		
	}

}